/**
 * @file
 * vuo.shader.make.shadertoy node implementation.
 *
 * @copyright Copyright © 2012–2014 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include "VuoGlContext.h"
#include "../vuo.mouse/VuoMouse.h"
#include <OpenGL/CGLMacro.h>

VuoModuleMetadata({
					 "title" : "Shade with Shadertoy",
					 "keywords" : [ "opengl", "glsl", "scenegraph", "graphics",
						 "lighting", "lit", "lighted",
						 "Blinn", "Phong", "Lambert", "tone", "chroma" ],
					 "version" : "2.0.0",
					 "dependencies" : [
						 "VuoMouse",
						 "VuoGlContext"
					 ],
					 "node" : {
						  "exampleCompositions" : [ ]
					 }
				 });

/**
 *	Shadertoy specific inputs
 *
 *	vec3		iResolution				image			The viewport resolution (z is pixel aspect ratio, usually 1.0)
 *	float		iGlobalTime				image/sound		Current time in seconds
 *	float		iChannelTime[4]			image			Time for channel (if video or sound), in seconds		!!!! NOT SUPPORTED
 *	vec3		iChannelResolution0..3	image/sound		Input texture resolution for each channel
 *	vec4		iMouse					image			xy = current pixel coords (if LMB is down). zw = click pixel
 *	sampler2D	iChannel{i}				image/sound		Sampler for input textures i
 *	vec4		iDate					image/sound		Year, month, day, time in seconds in .xyzw
 *	float		iSampleRate				image/sound		The sound sample rate (typically 44100)	-- In Vuo, const 44100 until AudioSample input is supported.  Only provided so that shaders will successfully compile.
 */

const char* ShaderHeader = "					\
#version 120								\n 	\
uniform vec3      iResolution;           	\n 	\
uniform float     iGlobalTime;           	\n 	\
uniform vec4      iMouse;					\n 	\
uniform float 	  iChannelTime[4]; 			\n  \
uniform vec3      iChannelResolution[4];	\n 	\
uniform sampler2D iChannel0;	         	\n 	\
uniform sampler2D iChannel1;	         	\n 	\
uniform sampler2D iChannel2;	         	\n 	\
uniform sampler2D iChannel3;	         	\n 	\
uniform vec4 	  iDate;					\n 	\
const float 	  iSampleRate = 44100;      \n  \
\n";

struct nodeInstanceData
{
	VuoShader shader;
	VuoPoint3d windowDimensions;
	VuoWindowReference window;

	VuoPoint2d clickedMousePosition;
	VuoPoint2d draggingMousePosition;

	VuoMouse *leftMouseDraggedListener;
	VuoMouse *mousePressedListener;
	VuoMouse *mouseReleasedListener;
};

/**
 *	Update nodeInstanceData's cached window and mouse trackers
 */
void updateWindow(struct nodeInstanceData* instance)
{
	int64_t pixelsWide, pixelsHigh;
	float backingScaleFactor = 1;

	if (!instance->window)
		VuoMouse_GetScreenDimensions(&pixelsWide, &pixelsHigh);
	else
		VuoWindowReference_getContentSize(instance->window, &pixelsWide, &pixelsHigh, &backingScaleFactor);

	instance->windowDimensions = VuoPoint3d_make((float)pixelsWide, (float)pixelsHigh, 1.);
}


struct nodeInstanceData * nodeInstanceInit(
	VuoInputData(VuoText) fragmentShader,
	VuoInputData(VuoWindowReference) window
	)
{
	struct nodeInstanceData * instance = (struct nodeInstanceData *)malloc(sizeof(struct nodeInstanceData));
	VuoRegister(instance, free);

	instance->shader = VuoShader_make("Shadertoy Fragment Shader");
	VuoRetain(instance->shader);

	char* fragmentSource = (char*)malloc(strlen(ShaderHeader) + strlen(fragmentShader) + 1);

	strcpy(fragmentSource, ShaderHeader);
	strcat(fragmentSource, fragmentShader);

	// VLog("\n\n=============================\n\n%s\n\n=============================\n", fragmentSource);

	VuoShader_addSource(instance->shader, VuoMesh_IndividualTriangles, NULL, NULL, fragmentSource);

	free(fragmentSource);

	instance->windowDimensions = VuoPoint3d_make(0., 0., 0.);
	instance->clickedMousePosition = VuoPoint2d_make(0., 0.);
	instance->draggingMousePosition = VuoPoint2d_make(0., 0.);

	instance->leftMouseDraggedListener = VuoMouse_make();
	instance->mousePressedListener = VuoMouse_make();
	instance->mouseReleasedListener = VuoMouse_make();

	VuoRetain(instance->leftMouseDraggedListener);
	VuoRetain(instance->mousePressedListener);
	VuoRetain(instance->mouseReleasedListener);

	instance->window = window;
	updateWindow(instance);

	return instance;
}

/**
 *	[wip and unused]
 *  Convert an audio samples array to a 2xSampleCount image, where the first row of pixels represents an FFT in grayscale, and the second
 *  is the normalized PCM data.
 */
VuoImage convertAudioToImage(VuoAudioSamples audio)
{
	int len = audio.sampleCount;

	unsigned char* pixels = (unsigned char*)malloc(sizeof(char)*len*4);
	int n = 0;
	for(int i = 0; i < len; i++)
	{
		pixels[n++] = (unsigned int)(audio.samples[i]*255);
		pixels[n++] = (unsigned int)(audio.samples[i]*255);
		pixels[n++] = (unsigned int)(audio.samples[i]*255);
		pixels[n++] = (unsigned int)(1.);
	}

	VuoImage audioImage = VuoImage_makeFromBuffer(pixels, GL_RGBA, len, 1, VuoImageColorDepth_8);

	return audioImage;
}

VuoPoint2d convertToPixelCoordinate(VuoPoint2d point, VuoPoint2d window)
{
	VuoPoint2d normalized;

	normalized.x = (point.x+1)/2.;
	float aspect = window.y / window.x;
	normalized.y = (point.y+aspect) / (aspect*2);

	return VuoPoint2d_scale(normalized, window);
}

void onLeftMouseDraggedSt(struct nodeInstanceData *instance, VuoPoint2d point)
{
	instance->draggingMousePosition = instance->window ? convertToPixelCoordinate(point, (VuoPoint2d){instance->windowDimensions.x, instance->windowDimensions.y}) : point;
}

void onMousePressedSt(struct nodeInstanceData *instance, VuoPoint2d point)
{
	instance->clickedMousePosition = instance->window ? convertToPixelCoordinate(point, (VuoPoint2d){instance->windowDimensions.x, instance->windowDimensions.y}) : point;
}

void onMouseReleasedSt(struct nodeInstanceData *instance, VuoPoint2d point)
{
	instance->clickedMousePosition = VuoPoint2d_make(0., 0.);
}

/**
 *	Begin listening with modifier key and window.
 */
void startMouseListenersSt(struct nodeInstanceData* instance)
{
	VuoMouse_startListeningForDragsWithCallback(instance->leftMouseDraggedListener,
											^(VuoPoint2d point) { onLeftMouseDraggedSt(instance, point); },
											VuoMouseButton_Left,
											instance->window,
											VuoModifierKey_None);

	VuoMouse_startListeningForPressesWithCallback(instance->mousePressedListener,
											^(VuoPoint2d point) { onMousePressedSt(instance, point); },
											VuoMouseButton_Any,
											instance->window,
											VuoModifierKey_None
											);

	VuoMouse_startListeningForReleasesWithCallback(instance->mouseReleasedListener,
											^(VuoPoint2d point) { onMouseReleasedSt(instance, point); },
											VuoMouseButton_Any,
											instance->window,
											VuoModifierKey_None);

}

void stopMouseListenersSt(struct nodeInstanceData* instance)
{
	VuoMouse_stopListening(instance->leftMouseDraggedListener);
	VuoMouse_stopListening(instance->mousePressedListener);
	VuoMouse_stopListening(instance->mouseReleasedListener);
}

void nodeInstanceEvent
(
		VuoInstanceData(struct nodeInstanceData *) instance,
		VuoInputData(VuoText, {
				"default": "void main(void)\n{\n	vec2 uv = (gl_FragCoord.xy/iResolution.xy);\n	gl_FragColor = vec4(uv.xyx, 1.);\n}"}
			) fragmentShader,
		VuoInputEvent({"eventBlocking":"none","data":"fragmentShader"}) fragmentShaderEvent,
		VuoInputData(VuoReal) GlobalTime,
		VuoInputData(VuoImage) Channel0,
		VuoInputData(VuoImage) Channel1,
		VuoInputData(VuoImage) Channel2,
		VuoInputData(VuoImage) Channel3,
		VuoInputData(VuoWindowReference) window,
		VuoInputEvent({"data":"window"}) windowEvent,
		VuoOutputData(VuoShader) shader
)
{
	/**
	 *	If the fragment shader text has changed, reload the shader.
	 */
	if(fragmentShaderEvent)
	{
		if((*instance)->shader)
		{
			VuoRelease((*instance)->shader);
		}

		(*instance)->shader = VuoShader_make("Shadertoy Fragment Shader");
		VuoRetain((*instance)->shader);

		char* fragmentSource = (char*)malloc(strlen(ShaderHeader) + strlen(fragmentShader) + 1);

		strcpy(fragmentSource, ShaderHeader);
		strcat(fragmentSource, fragmentShader);

		VuoShader_addSource((*instance)->shader, VuoMesh_IndividualTriangles, NULL, NULL, fragmentSource);

		free(fragmentSource);
	}

	/**
	 *	If the window reference has changed or updated, reset the resolution and mouse tracker.
	 */
	if(windowEvent)
	{
		stopMouseListenersSt(*instance);

		(*instance)->window = window;
		updateWindow(*instance);

		startMouseListenersSt(*instance);
	}

	/**
	 *	Time
	 */
	VuoShader_setUniform_VuoReal((*instance)->shader, "iGlobalTime", GlobalTime);

	/**
	 *	Window resolution - does not have to be provided.
	 */
	VuoShader_setUniform_VuoPoint3d((*instance)->shader, "iResolution", (*instance)->windowDimensions);

	/**
	 *	Mouse Coordinates / Clicked
	 */
	VuoShader_setUniform_VuoPoint4d( (*instance)->shader, "iMouse", VuoPoint4d_make(
		(*instance)->draggingMousePosition.x,
		(*instance)->draggingMousePosition.y,
		(*instance)->clickedMousePosition.x,
		(*instance)->clickedMousePosition.y) );

	/**
	 *	Channels (Textures and Audio ...eventually)
	 */
	if(Channel0 != NULL)
	{
		VuoShader_setUniform_VuoPoint3d((*instance)->shader, "iChannelResolution[0]", VuoPoint3d_make(Channel0->pixelsWide, Channel0->pixelsHigh, 1.0));
		VuoShader_setUniform_VuoImage((*instance)->shader, "iChannel0", Channel0);
	}

	if(Channel1 != NULL)
	{
		VuoShader_setUniform_VuoPoint3d((*instance)->shader, "iChannelResolution[1]", VuoPoint3d_make(Channel1->pixelsWide, Channel1->pixelsHigh, 1.0));
		VuoShader_setUniform_VuoImage((*instance)->shader, "iChannel1", Channel1);
	}

	if(Channel2 != NULL)
	{
		VuoShader_setUniform_VuoPoint3d((*instance)->shader, "iChannelResolution[2]", VuoPoint3d_make(Channel2->pixelsWide, Channel2->pixelsHigh, 1.0));
		VuoShader_setUniform_VuoImage((*instance)->shader, "iChannel2", Channel2);
	}

	if(Channel3 != NULL)
	{
		VuoShader_setUniform_VuoPoint3d((*instance)->shader, "iChannelResolution[3]", VuoPoint3d_make(Channel3->pixelsWide, Channel3->pixelsHigh, 1.0));
		VuoShader_setUniform_VuoImage((*instance)->shader, "iChannel3", Channel3);
	}

	/**
	 *	Date and time.
	 * 	Blame this guy if not correct: http://stackoverflow.com/questions/1442116/how-to-get-date-and-time-value-in-c-program
	 */
	time_t t = time(NULL);
	struct tm tm = *localtime(&t);
	VuoShader_setUniform_VuoPoint4d((*instance)->shader, "iDate", (VuoPoint4d) { tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_sec } );

	*shader = (*instance)->shader;
}

void nodeInstanceTriggerStart(
	VuoInstanceData(struct nodeInstanceData *) instance,
	VuoInputData(VuoWindowReference) window
	)
{
	(*instance)->window = window;

	updateWindow(*instance);

	startMouseListenersSt(*instance);
}

void nodeInstanceTriggerStop(VuoInstanceData(struct nodeInstanceData *) instance)
{
		stopMouseListenersSt(*instance);
}

void nodeInstanceFini(VuoInstanceData(struct nodeInstanceData *) instance)
{
	VuoRelease((*instance)->leftMouseDraggedListener);
	VuoRelease((*instance)->mousePressedListener);
	VuoRelease((*instance)->mouseReleasedListener);

	VuoRelease((*instance)->shader);
}
